@@ -232,7 +232,7 @@ require AutoLoader;
        LDAP_TAG_IM_RES_VALUE
        LDAP_TAG_SASL_RES_CREDS
        );
-$VERSION = '3.0.4';
+$VERSION = '3.0.x';
 
 sub AUTOLOAD {
     # This AUTOLOAD is used to 'autoload' constants from the constant()
@@ -296,7 +296,10 @@ sub new
     syncCookie ::= OCTET STRING
 
     syncRequestValue ::= SEQUENCE {
-        mode       ENUMERATED,
+				mode ENUMERATED {
+						refreshOnly       (1),
+						refreshAndPersist (3)
+				},
         cookie     syncCookie OPTIONAL,
         reloadHint BOOLEAN
         }
@@ -881,10 +884,7 @@ sub next_changed_entries {
                 $cookie = $syncStateValues->{'cookie'};
                 if( $cookie ) {
                     # save the cookie
-                    open(COOKIE_FILE,">".$self->{"cookie"}) ||
-                        die("Cannot open file '".$self->{"cookie"}."' for writing.");
-                    print COOKIE_FILE $cookie;
-                    close(COOKIE_FILE);
+										save_cookie($cookie, $self->{"cookie"});
                 }
             }
 
@@ -916,10 +916,7 @@ sub next_changed_entries {
 
                 # see if we got any and save it.
                 if( $cookie ) {
-                    open(COOKIE_FILE,">".$self->{"cookie"}) ||
-                        die("Cannot open file '".$self->{"cookie"}."' for writing.");
-                    print COOKIE_FILE $cookie;
-                    close(COOKIE_FILE);
+										save_cookie($cookie, $self->{"cookie"});
                 }
             }
         }
@@ -928,6 +925,52 @@ sub next_changed_entries {
     return @entries;
 } # next_changed_entries
 
+sub save_cookie
+{
+    my ($self,@args) = @_;
+    my $cookiestr = $_[0];
+    my $cookie = $_[1];
+
+    # Skip all if there's no csn value
+    if ($cookiestr =~ m/csn=/) {
+
+        # Get new CSN array and a copy
+        chomp(my @newcsns = split(';',$cookiestr =~ s/(rid=\d{3},)|(sid=\d{3},)|(csn=)//rg));
+
+        # These will be the CSNs to write to the cookie file
+        # All CSNs from the new cookie must be used
+        # my @outcsns = @newcsns;
+        my @outcsns = @newcsns;
+
+        # Get the old cookie for comparison/persisting
+        if (-w $cookie) {
+            open(COOKIE_FILE, "<", $cookie) || die("Cannot open file '".$cookie."' for reading.");
+            chomp(my @oldcsns = <COOKIE_FILE>);
+            close(COOKIE_FILE);
+
+            # Look for old CSNs with SIDs that don't match any of the new
+            # CSNs. If there are no matches, push the old CSN to the list
+            # of CSNs to be written to the cookie file.
+            foreach my $oldcsn (@oldcsns) {
+                my $match = 0;
+                my $p_sid  = ($oldcsn =~ /(#\d{3}#)/i)[0];
+                foreach my $newcsn (@newcsns) {
+                    if ($newcsn =~ m/\Q$p_sid/) {
+                        $match = 1;
+                        last;
+                    }
+                }
+                if (!$match) { push @outcsns,$oldcsn; }
+            }
+        }
+
+        # Write the cookie
+        open(COOKIE_FILE, ">", $cookie) || die("Cannot open file '".$cookie."' for writing.");
+        print COOKIE_FILE "$_\n" for @outcsns;
+        close(COOKIE_FILE);
+    }
+} # end save_cookie
+
 
 sub first_entry
 {
@@ -1529,14 +1572,15 @@ sub listen_for_changes
     my ($msgid, $status, $sctrls, $the_cookie, $syncRequestBerval);
 
     my ($basedn,    $scope,   $filter,    $attrs,
-        $attrsonly, $timeout, $sizelimit, $cookie) =
+        $attrsonly, $timeout, $sizelimit, $cookie, $rid) =
             $self->rearrange(['BASEDN',    'SCOPE',   'FILTER',    'ATTRS',
-                              'ATTRSONLY', 'TIMEOUT', 'SIZELIMIT', 'COOKIE'], @args);
+                              'ATTRSONLY', 'TIMEOUT', 'SIZELIMIT', 'COOKIE', 'RID'], @args);
 
     croak("No Filter Specified") if (!defined($filter));
     croak("No cookie file specified") unless $cookie;
 
     $self->{"cookie"} = $cookie;
+    $self->{"rid"} = defined($rid) ? $rid : '000';
 
     if( !defined($attrs) ) {
         my @null_array = ();
@@ -1545,7 +1589,10 @@ sub listen_for_changes
 
     # load cookie from the file
     if( open(COOKIE, $cookie) ) {
-        read COOKIE, $the_cookie, 1024, 0;
+				chomp(my @csns = <COOKIE>);
+        if (scalar(@csns)) {
+				  $the_cookie = sprintf("rid=%d,csn=%s",$rid,join(';',@csns));
+        }
     } else {
         warn "Failed to open file '".$cookie."' for reading.\n";
     }
@@ -1555,9 +1602,9 @@ sub listen_for_changes
 
     # refreshAndPersist mode
     if( $the_cookie ) { # we have the cookie
-        $syncRequestBerval = $asn->encode(mode => 3, cookie => $the_cookie, reloadHint => 1);
+        $syncRequestBerval = $syncRequestValue->encode(mode => 3, cookie => $the_cookie, reloadHint => 1);
     } else {
-        $syncRequestBerval = $asn->encode(mode => 3, reloadHint => 1);
+        $syncRequestBerval = $syncRequestValue->encode(mode => 3, reloadHint => 1);
     }
 
     my $ctrl_persistent =
@@ -2468,6 +2515,8 @@ Net::LDAPapi - Perl5 Module Supporting LDAP API
            
 =head1 SUPPORTED METHODS
 
+=over 4
+
 =item abandon MSGID SCTRLS CCTRLS
 
   This cancels an asynchronous LDAP operation that has not completed.  It
